#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=unused-import,reimported,import-error,no-name-in-module
import unittest


from neptune.api_exceptions import ChannelAlreadyExists
from neptune.api_exceptions import ChannelDoesNotExist
from neptune.api_exceptions import ChannelNotFound
from neptune.api_exceptions import ChannelsValuesSendBatchError
from neptune.api_exceptions import ConnectionLost
from neptune.api_exceptions import ExperimentAlreadyFinished
from neptune.api_exceptions import ExperimentLimitReached
from neptune.api_exceptions import ExperimentNotFound
from neptune.api_exceptions import ExperimentOperationErrors
from neptune.api_exceptions import ExperimentValidationError
from neptune.api_exceptions import Forbidden
from neptune.api_exceptions import InvalidApiKey
from neptune.api_exceptions import NeptuneApiException
from neptune.api_exceptions import NeptuneException
from neptune.api_exceptions import NotebookNotFound
from neptune.api_exceptions import PathInExperimentNotFound
from neptune.api_exceptions import PathInProjectNotFound
from neptune.api_exceptions import ProjectNotFound
from neptune.api_exceptions import SSLError
from neptune.api_exceptions import ServerError
from neptune.api_exceptions import StorageLimitReached
from neptune.api_exceptions import Unauthorized
from neptune.api_exceptions import WorkspaceNotFound
from neptune.backend import ABC
from neptune.backend import ApiClient
from neptune.backend import BackendApiClient
from neptune.backend import ChannelWithLastValue
from neptune.backend import Dict
from neptune.backend import LeaderboardApiClient
from neptune.checkpoint import Checkpoint
from neptune.exceptions import CannotResolveHostname
from neptune.exceptions import DeleteArtifactUnsupportedInAlphaException
from neptune.exceptions import DeprecatedApiToken
from neptune.exceptions import DownloadArtifactUnsupportedException
from neptune.exceptions import DownloadArtifactsUnsupportedException
from neptune.exceptions import DownloadSourcesException
from neptune.exceptions import FileNotFound
from neptune.exceptions import InvalidChannelValue
from neptune.exceptions import InvalidNeptuneBackend
from neptune.exceptions import InvalidNotebookPath
from neptune.exceptions import NeptuneException
from neptune.exceptions import NeptuneIncorrectImportException
from neptune.exceptions import NeptuneIncorrectProjectQualifiedNameException
from neptune.exceptions import NeptuneLibraryNotInstalledException
from neptune.exceptions import NeptuneMissingApiTokenException
from neptune.exceptions import NeptuneMissingProjectQualifiedNameException
from neptune.exceptions import NeptuneNoExperimentContextException
from neptune.exceptions import NeptuneUninitializedException
from neptune.exceptions import NoChannelValue
from neptune.exceptions import NotADirectory
from neptune.exceptions import NotAFile
from neptune.exceptions import ProjectMigratedToNewStructure
from neptune.exceptions import UnsupportedClientVersion
from neptune.exceptions import UnsupportedInAlphaException
from neptune.experiments import ChannelDoesNotExist
from neptune.experiments import ChannelNamespace
from neptune.experiments import ChannelType
from neptune.experiments import ChannelValue
from neptune.experiments import ChannelsValuesSender
from neptune.experiments import EmptyDataError
from neptune.experiments import ExecutionContext
from neptune.experiments import Experiment
from neptune.experiments import ExperimentAlreadyFinished
from neptune.experiments import InvalidChannelValue
from neptune.experiments import NeptuneIncorrectImportException
from neptune.experiments import NoChannelValue
from neptune.git_info import GitInfo
from neptune.management.exceptions import AccessRevokedOnDeletion
from neptune.management.exceptions import AccessRevokedOnMemberRemoval
from neptune.management.exceptions import BadRequestException
from neptune.management.exceptions import ConflictingWorkspaceName
from neptune.management.exceptions import InvalidProjectName
from neptune.management.exceptions import ManagementOperationFailure
from neptune.management.exceptions import MissingWorkspaceName
from neptune.management.exceptions import ProjectAlreadyExists
from neptune.management.exceptions import ProjectNotFound
from neptune.management.exceptions import ProjectsLimitReached
from neptune.management.exceptions import UnsupportedValue
from neptune.management.exceptions import UserAlreadyHasAccess
from neptune.management.exceptions import UserNotExistsOrWithoutAccess
from neptune.management.exceptions import WorkspaceNotFound
from neptune.model import ChannelWithLastValue
from neptune.model import LeaderboardEntry
from neptune.model import Point
from neptune.model import Points
from neptune.new.attributes.atoms.artifact import Artifact
from neptune.new.attributes.atoms.artifact import ArtifactDriver
from neptune.new.attributes.atoms.artifact import ArtifactDriversMap
from neptune.new.attributes.atoms.artifact import ArtifactFileData
from neptune.new.attributes.atoms.artifact import ArtifactVal
from neptune.new.attributes.atoms.artifact import AssignArtifact
from neptune.new.attributes.atoms.artifact import Atom
from neptune.new.attributes.atoms.artifact import OptionalFeatures
from neptune.new.attributes.atoms.artifact import TrackFilesToArtifact
from neptune.new.attributes.atoms.atom import Atom
from neptune.new.attributes.atoms.atom import Attribute
from neptune.new.attributes.atoms.boolean import AssignBool
from neptune.new.attributes.atoms.boolean import Boolean
from neptune.new.attributes.atoms.boolean import BooleanVal
from neptune.new.attributes.atoms.datetime import AssignDatetime
from neptune.new.attributes.atoms.datetime import Datetime
from neptune.new.attributes.atoms.datetime import DatetimeVal
from neptune.new.attributes.atoms.datetime import datetime
from neptune.new.attributes.atoms.file import Atom
from neptune.new.attributes.atoms.file import File
from neptune.new.attributes.atoms.file import FileVal
from neptune.new.attributes.atoms.file import UploadFile
from neptune.new.attributes.atoms.file import UploadFileContent
from neptune.new.attributes.atoms.float import AssignFloat
from neptune.new.attributes.atoms.float import Float
from neptune.new.attributes.atoms.float import FloatVal
from neptune.new.attributes.atoms.git_ref import Atom
from neptune.new.attributes.atoms.git_ref import GitRef
from neptune.new.attributes.atoms.integer import AssignInt
from neptune.new.attributes.atoms.integer import Integer
from neptune.new.attributes.atoms.integer import IntegerVal
from neptune.new.attributes.atoms.notebook_ref import Atom
from neptune.new.attributes.atoms.notebook_ref import NotebookRef
from neptune.new.attributes.atoms.run_state import Atom
from neptune.new.attributes.atoms.run_state import RunState
from neptune.new.attributes.atoms.string import AssignString
from neptune.new.attributes.atoms.string import String
from neptune.new.attributes.atoms.string import StringVal
from neptune.new.attributes.attribute import Attribute
from neptune.new.attributes.attribute import List
from neptune.new.attributes.attribute import NeptuneBackend
from neptune.new.attributes.attribute import Operation
from neptune.new.attributes.file_set import Attribute
from neptune.new.attributes.file_set import DeleteFiles
from neptune.new.attributes.file_set import FileSet
from neptune.new.attributes.file_set import FileSetVal
from neptune.new.attributes.file_set import Iterable
from neptune.new.attributes.file_set import UploadFileSet
from neptune.new.attributes.namespace import Attribute
from neptune.new.attributes.namespace import Dict
from neptune.new.attributes.namespace import Iterator
from neptune.new.attributes.namespace import List
from neptune.new.attributes.namespace import Mapping
from neptune.new.attributes.namespace import MutableMapping
from neptune.new.attributes.namespace import Namespace
from neptune.new.attributes.namespace import NamespaceBuilder
from neptune.new.attributes.namespace import NamespaceVal
from neptune.new.attributes.namespace import NoValue
from neptune.new.attributes.namespace import RunStructure
from neptune.new.attributes.series.fetchable_series import Dict
from neptune.new.attributes.series.fetchable_series import FetchableSeries
from neptune.new.attributes.series.fetchable_series import FloatSeriesValues
from neptune.new.attributes.series.fetchable_series import Generic
from neptune.new.attributes.series.fetchable_series import StringSeriesValues
from neptune.new.attributes.series.fetchable_series import TypeVar
from neptune.new.attributes.series.fetchable_series import datetime
from neptune.new.attributes.series.file_series import ClearImageLog
from neptune.new.attributes.series.file_series import Data
from neptune.new.attributes.series.file_series import File
from neptune.new.attributes.series.file_series import FileNotFound
from neptune.new.attributes.series.file_series import FileSeries
from neptune.new.attributes.series.file_series import FileSeriesVal
from neptune.new.attributes.series.file_series import ImageValue
from neptune.new.attributes.series.file_series import Iterable
from neptune.new.attributes.series.file_series import LogImages
from neptune.new.attributes.series.file_series import Operation
from neptune.new.attributes.series.file_series import OperationNotSupported
from neptune.new.attributes.series.file_series import Series
from neptune.new.attributes.series.file_series import Val
from neptune.new.attributes.series.float_series import ClearFloatLog
from neptune.new.attributes.series.float_series import ConfigFloatSeries
from neptune.new.attributes.series.float_series import FetchableSeries
from neptune.new.attributes.series.float_series import FloatSeries
from neptune.new.attributes.series.float_series import FloatSeriesVal
from neptune.new.attributes.series.float_series import FloatSeriesValues
from neptune.new.attributes.series.float_series import Iterable
from neptune.new.attributes.series.float_series import LogFloats
from neptune.new.attributes.series.float_series import Operation
from neptune.new.attributes.series.float_series import Series
from neptune.new.attributes.series.float_series import Val
from neptune.new.attributes.series.series import Attribute
from neptune.new.attributes.series.series import Generic
from neptune.new.attributes.series.series import Iterable
from neptune.new.attributes.series.series import Operation
from neptune.new.attributes.series.series import Series
from neptune.new.attributes.series.series import SeriesVal
from neptune.new.attributes.series.series import TypeVar
from neptune.new.attributes.series.string_series import ClearStringLog
from neptune.new.attributes.series.string_series import Data
from neptune.new.attributes.series.string_series import FetchableSeries
from neptune.new.attributes.series.string_series import Iterable
from neptune.new.attributes.series.string_series import List
from neptune.new.attributes.series.string_series import LogStrings
from neptune.new.attributes.series.string_series import Operation
from neptune.new.attributes.series.string_series import Series
from neptune.new.attributes.series.string_series import StringSeries
from neptune.new.attributes.series.string_series import StringSeriesVal
from neptune.new.attributes.series.string_series import StringSeriesValues
from neptune.new.attributes.series.string_series import Val
from neptune.new.attributes.sets.set import Attribute
from neptune.new.attributes.sets.set import Set
from neptune.new.attributes.sets.string_set import AddStrings
from neptune.new.attributes.sets.string_set import ClearStringSet
from neptune.new.attributes.sets.string_set import Iterable
from neptune.new.attributes.sets.string_set import RemoveStrings
from neptune.new.attributes.sets.string_set import Set
from neptune.new.attributes.sets.string_set import StringSet
from neptune.new.attributes.sets.string_set import StringSetVal
from neptune.new.attributes.utils import Artifact
from neptune.new.attributes.utils import AttributeType
from neptune.new.attributes.utils import Boolean
from neptune.new.attributes.utils import Datetime
from neptune.new.attributes.utils import File
from neptune.new.attributes.utils import FileSeries
from neptune.new.attributes.utils import FileSet
from neptune.new.attributes.utils import Float
from neptune.new.attributes.utils import FloatSeries
from neptune.new.attributes.utils import GitRef
from neptune.new.attributes.utils import Integer
from neptune.new.attributes.utils import InternalClientError
from neptune.new.attributes.utils import List
from neptune.new.attributes.utils import NotebookRef
from neptune.new.attributes.utils import RunState
from neptune.new.attributes.utils import String
from neptune.new.attributes.utils import StringSeries
from neptune.new.attributes.utils import StringSet
from neptune.new.exceptions import ArtifactNotFoundException
from neptune.new.exceptions import ArtifactUploadingError
from neptune.new.exceptions import CannotResolveHostname
from neptune.new.exceptions import CannotSynchronizeOfflineRunsWithoutProject
from neptune.new.exceptions import ClientHttpError
from neptune.new.exceptions import ExceptionWithProjectsWorkspacesListing
from neptune.new.exceptions import FetchAttributeNotFoundException
from neptune.new.exceptions import FileNotFound
from neptune.new.exceptions import FileSetUploadError
from neptune.new.exceptions import FileUploadError
from neptune.new.exceptions import Forbidden
from neptune.new.exceptions import InactiveRunException
from neptune.new.exceptions import InternalClientError
from neptune.new.exceptions import InternalServerError
from neptune.new.exceptions import List
from neptune.new.exceptions import MalformedOperation
from neptune.new.exceptions import MetadataInconsistency
from neptune.new.exceptions import MissingFieldException
from neptune.new.exceptions import NeedExistingRunForReadOnlyMode
from neptune.new.exceptions import NeptuneApiException
from neptune.new.exceptions import NeptuneConnectionLostException
from neptune.new.exceptions import NeptuneEmptyLocationException
from neptune.new.exceptions import NeptuneException
from neptune.new.exceptions import NeptuneFeatureNotAvailableException
from neptune.new.exceptions import NeptuneIntegrationNotInstalledException
from neptune.new.exceptions import NeptuneInvalidApiTokenException
from neptune.new.exceptions import NeptuneLegacyIncompatibilityException
from neptune.new.exceptions import NeptuneLegacyProjectException
from neptune.new.exceptions import NeptuneLimitExceedException
from neptune.new.exceptions import NeptuneLocalStorageAccessException
from neptune.new.exceptions import NeptuneMissingApiTokenException
from neptune.new.exceptions import NeptuneMissingProjectNameException
from neptune.new.exceptions import NeptuneOfflineModeFetchException
from neptune.new.exceptions import NeptunePossibleLegacyUsageException
from neptune.new.exceptions import NeptuneRemoteStorageAccessException
from neptune.new.exceptions import NeptuneRemoteStorageCredentialsException
from neptune.new.exceptions import NeptuneRunResumeAndCustomIdCollision
from neptune.new.exceptions import NeptuneStorageLimitException
from neptune.new.exceptions import NeptuneUnhandledArtifactSchemeException
from neptune.new.exceptions import NeptuneUnhandledArtifactTypeException
from neptune.new.exceptions import NeptuneUninitializedException
from neptune.new.exceptions import NeptuneUnsupportedArtifactFunctionalityException
from neptune.new.exceptions import OperationNotSupported
from neptune.new.exceptions import PlotlyIncompatibilityException
from neptune.new.exceptions import Project
from neptune.new.exceptions import ProjectNameCollision
from neptune.new.exceptions import ProjectNotFound
from neptune.new.exceptions import RunNotFound
from neptune.new.exceptions import RunUUIDNotFound
from neptune.new.exceptions import SSLError
from neptune.new.exceptions import Unauthorized
from neptune.new.exceptions import UnsupportedClientVersion
from neptune.new.exceptions import Version
from neptune.new.exceptions import Workspace
from neptune.new.handler import Artifact
from neptune.new.handler import ArtifactFileData
from neptune.new.handler import File
from neptune.new.handler import FileSeries
from neptune.new.handler import FileSet
from neptune.new.handler import FileVal
from neptune.new.handler import FloatSeries
from neptune.new.handler import Handler
from neptune.new.handler import Iterable
from neptune.new.handler import List
from neptune.new.handler import NeptuneException
from neptune.new.handler import StringSeries
from neptune.new.handler import StringSet
from neptune.new.integrations.python_logger import Logger
from neptune.new.integrations.python_logger import NeptuneHandler
from neptune.new.integrations.python_logger import Run
from neptune.new.integrations.python_logger import RunState
from neptune.new.logging.logger import Logger
from neptune.new.logging.logger import Run
from neptune.new.project import Iterable
from neptune.new.project import Project
from neptune.new.project import RunsTable
from neptune.new.run import Attribute
from neptune.new.run import AttributeType
from neptune.new.run import BackgroundJob
from neptune.new.run import Boolean
from neptune.new.run import Datetime
from neptune.new.run import DeleteAttribute
from neptune.new.run import Float
from neptune.new.run import Handler
from neptune.new.run import InactiveRunException
from neptune.new.run import Integer
from neptune.new.run import MetadataInconsistency
from neptune.new.run import Namespace
from neptune.new.run import NamespaceAttr
from neptune.new.run import NamespaceBuilder
from neptune.new.run import NeptuneBackend
from neptune.new.run import NeptunePossibleLegacyUsageException
from neptune.new.run import OperationProcessor
from neptune.new.run import Run
from neptune.new.run import RunState
from neptune.new.run import RunStructure
from neptune.new.run import String
from neptune.new.run import Value
from neptune.new.run import ValueToAttributeVisitor
from neptune.new.runs_table import AttributeType
from neptune.new.runs_table import AttributeWithProperties
from neptune.new.runs_table import Dict
from neptune.new.runs_table import LeaderboardEntry
from neptune.new.runs_table import LeaderboardHandler
from neptune.new.runs_table import List
from neptune.new.runs_table import MetadataInconsistency
from neptune.new.runs_table import RunsTable
from neptune.new.runs_table import RunsTableEntry
from neptune.new.runs_table import datetime
from neptune.new.sync import ApiRun
from neptune.new.sync import CannotSynchronizeOfflineRunsWithoutProject
from neptune.new.sync import Credentials
from neptune.new.sync import DiskQueue
from neptune.new.sync import HostedNeptuneBackend
from neptune.new.sync import Iterable
from neptune.new.sync import List
from neptune.new.sync import NeptuneBackend
from neptune.new.sync import NeptuneConnectionLostException
from neptune.new.sync import NeptuneException
from neptune.new.sync import Operation
from neptune.new.sync import Path
from neptune.new.sync import Project
from neptune.new.sync import ProjectNotFound
from neptune.new.sync import RunNotFound
from neptune.new.sync import Sequence
from neptune.new.sync import Tuple
from neptune.new.types.atoms.artifact import Artifact
from neptune.new.types.atoms.artifact import Atom
from neptune.new.types.atoms.artifact import FileHasher
from neptune.new.types.atoms.artifact import TypeVar
from neptune.new.types.atoms.atom import Atom
from neptune.new.types.atoms.atom import TypeVar
from neptune.new.types.atoms.atom import Value
from neptune.new.types.atoms.boolean import Atom
from neptune.new.types.atoms.boolean import Boolean
from neptune.new.types.atoms.boolean import TypeVar
from neptune.new.types.atoms.datetime import Atom
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.datetime import TypeVar
from neptune.new.types.atoms.datetime import datetime
from neptune.new.types.atoms.file import Atom
from neptune.new.types.atoms.file import File
from neptune.new.types.atoms.file import IOBase
from neptune.new.types.atoms.file import TypeVar
from neptune.new.types.atoms.float import Atom
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.float import TypeVar
from neptune.new.types.atoms.git_ref import Atom
from neptune.new.types.atoms.git_ref import GitRef
from neptune.new.types.atoms.git_ref import List
from neptune.new.types.atoms.git_ref import TypeVar
from neptune.new.types.atoms.git_ref import datetime
from neptune.new.types.atoms.integer import Atom
from neptune.new.types.atoms.integer import Integer
from neptune.new.types.atoms.integer import TypeVar
from neptune.new.types.atoms.string import Atom
from neptune.new.types.atoms.string import String
from neptune.new.types.atoms.string import TypeVar
from neptune.new.types.file_set import FileSet
from neptune.new.types.file_set import Iterable
from neptune.new.types.file_set import List
from neptune.new.types.file_set import TypeVar
from neptune.new.types.file_set import Value
from neptune.new.types.namespace import Namespace
from neptune.new.types.namespace import TypeVar
from neptune.new.types.namespace import Value
from neptune.new.types.series.file_series import File
from neptune.new.types.series.file_series import FileSeries
from neptune.new.types.series.file_series import List
from neptune.new.types.series.file_series import Series
from neptune.new.types.series.file_series import TypeVar
from neptune.new.types.series.float_series import FloatSeries
from neptune.new.types.series.float_series import Series
from neptune.new.types.series.float_series import TypeVar
from neptune.new.types.series.series import Series
from neptune.new.types.series.series import TypeVar
from neptune.new.types.series.series import Value
from neptune.new.types.series.series_value import Generic
from neptune.new.types.series.series_value import SeriesValue
from neptune.new.types.series.series_value import TypeVar
from neptune.new.types.series.string_series import Series
from neptune.new.types.series.string_series import StringSeries
from neptune.new.types.series.string_series import TypeVar
from neptune.new.types.sets.set import Set
from neptune.new.types.sets.set import TypeVar
from neptune.new.types.sets.set import Value
from neptune.new.types.sets.string_set import Iterable
from neptune.new.types.sets.string_set import Set
from neptune.new.types.sets.string_set import StringSet
from neptune.new.types.sets.string_set import TypeVar
from neptune.new.types.value import TypeVar
from neptune.new.types.value import Value
from neptune.new.types.value_visitor import Artifact
from neptune.new.types.value_visitor import Boolean
from neptune.new.types.value_visitor import Datetime
from neptune.new.types.value_visitor import File
from neptune.new.types.value_visitor import FileSeries
from neptune.new.types.value_visitor import FileSet
from neptune.new.types.value_visitor import Float
from neptune.new.types.value_visitor import FloatSeries
from neptune.new.types.value_visitor import Generic
from neptune.new.types.value_visitor import GitRef
from neptune.new.types.value_visitor import Integer
from neptune.new.types.value_visitor import Namespace
from neptune.new.types.value_visitor import String
from neptune.new.types.value_visitor import StringSeries
from neptune.new.types.value_visitor import StringSet
from neptune.new.types.value_visitor import TypeVar
from neptune.new.types.value_visitor import Value
from neptune.new.types.value_visitor import ValueVisitor
from neptune.notebook import Notebook
from neptune.oauth import AuthBase
from neptune.oauth import Authenticator
from neptune.oauth import HTTPUnauthorized
from neptune.oauth import NeptuneAuth
from neptune.oauth import NeptuneAuthenticator
from neptune.oauth import NeptuneInvalidApiTokenException
from neptune.oauth import OAuth2Error
from neptune.oauth import OAuth2Session
from neptune.oauth import TokenExpiredError
from neptune.projects import DefaultAbortImpl
from neptune.projects import Experiment
from neptune.projects import NeptuneNoExperimentContextException
from neptune.projects import Project
from neptune.sessions import HostedNeptuneBackendApiClient
from neptune.sessions import OrderedDict
from neptune.sessions import Project
from neptune.sessions import Session
from neptune.utils import BravadoConnectionError
from neptune.utils import BravadoTimeoutError
from neptune.utils import ConnectionLost
from neptune.utils import FileNotFound
from neptune.utils import Forbidden
from neptune.utils import GitInfo
from neptune.utils import HTTPBadGateway
from neptune.utils import HTTPBadRequest
from neptune.utils import HTTPForbidden
from neptune.utils import HTTPGatewayTimeout
from neptune.utils import HTTPInternalServerError
from neptune.utils import HTTPRequestTimeout
from neptune.utils import HTTPServerError
from neptune.utils import HTTPServiceUnavailable
from neptune.utils import HTTPUnauthorized
from neptune.utils import InvalidNotebookPath
from neptune.utils import JSONDecodeError
from neptune.utils import NeptuneIncorrectProjectQualifiedNameException
from neptune.utils import NeptuneMissingProjectQualifiedNameException
from neptune.utils import NoopObject
from neptune.utils import NotADirectory
from neptune.utils import NotAFile
from neptune.utils import ProjectMigratedToNewStructure
from neptune.utils import SSLError
from neptune.utils import ServerError
from neptune.utils import Unauthorized


class TestImports(unittest.TestCase):
    def test_imports(self):
        pass



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os
from typing import Optional, List, Dict

from bravado.client import SwaggerClient
from bravado.exception import (
    HTTPNotFound,
    HTTPBadRequest,
    HTTPConflict,
    HTTPForbidden,
    HTTPUnprocessableEntity,
)

from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN
from neptune.new.envs import API_TOKEN_ENV_NAME
from neptune.new.internal.utils import verify_type
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.backends.hosted_client import (
    create_backend_client,
    create_http_client_with_auth,
    DEFAULT_REQUEST_KWARGS,
)
from neptune.new.internal.backends.utils import (
    with_api_exceptions_handler,
    ssl_verify,
    parse_validation_errors,
)
from neptune.management.internal.utils import normalize_project_name
from neptune.management.internal.types import *
from neptune.management.exceptions import (
    AccessRevokedOnDeletion,
    AccessRevokedOnMemberRemoval,
    ProjectAlreadyExists,
    ProjectNotFound,
    UserNotExistsOrWithoutAccess,
    WorkspaceNotFound,
    UserAlreadyHasAccess,
    BadRequestException,
    ProjectsLimitReached,
)
from neptune.management.internal.dto import (
    ProjectVisibilityDTO,
    ProjectMemberRoleDTO,
    WorkspaceMemberRoleDTO,
)


def _get_token(api_token: Optional[str] = None) -> str:
    return api_token or os.getenv(API_TOKEN_ENV_NAME)


def _get_backend_client(api_token: Optional[str] = None) -> SwaggerClient:
    credentials = Credentials.from_token(api_token=_get_token(api_token=api_token))
    http_client, client_config = create_http_client_with_auth(
        credentials=credentials, ssl_verify=ssl_verify(), proxies={}
    )
    return create_backend_client(client_config=client_config, http_client=http_client)


@with_api_exceptions_handler
def get_project_list(api_token: Optional[str] = None) -> List[str]:
    """Get a list of projects you have access to.
    Args:
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Returns:
        ``List[str]``: list of project names of a format 'WORKSPACE/PROJECT'
    Examples:
        >>> from neptune import management
        >>> management.get_project_list()
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    """
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)

    params = {
        "userRelation": "viewerOrHigher",
        "sortBy": ["lastViewed"],
        **DEFAULT_REQUEST_KWARGS,
    }

    projects = backend_client.api.listProjects(**params).response().result.entries

    return [
        normalize_project_name(name=project.name, workspace=project.organizationName)
        for project in projects
    ]


@with_api_exceptions_handler
def create_project(
    name: str,
    key: str,
    workspace: Optional[str] = None,
    visibility: str = ProjectVisibility.PRIVATE,
    description: Optional[str] = None,
    api_token: Optional[str] = None,
) -> str:
    """Creates a new project in your Neptune workspace.
    Args:
        name(str): The name of the project in Neptune in the format 'WORKSPACE/PROJECT'.
            If workspace argument was set, it should only contain 'PROJECT' instead of 'WORKSPACE/PROJECT'.
        key(str): Project identifier. It has to be contain 1-10 upper case letters or numbers.
            For example, 'GOOD5'
        workspace(str, optional): Name of your Neptune workspace.
            If you specify it, change the format of the name argument to 'PROJECT' instead of 'WORKSPACE/PROJECT'.
            If 'None' it will be parsed from the `name` argument.
        visibility(str, optional): level of visibility you want your project to have.
            Can be set to:
             - 'pub' for public projects
             - 'priv' for private projects
            If 'None' it will be set to 'priv'
        description(str, optional): Project description.
            If 'None', it will be left empty.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Returns:
        ``str``: name of the new project you created.
    Examples:
        >>> from neptune import management
        >>> management.create_project(name="awesome-team/amazing-project",
        ...                           key="AMA",
        ...                           visibility="pub")
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    """
    verify_type("name", name, str)
    verify_type("key", key, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("visibility", visibility, str)
    verify_type("description", description, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    project_spec = re.search(PROJECT_QUALIFIED_NAME_PATTERN, project_identifier)
    workspace, name = project_spec["workspace"], project_spec["project"]

    try:
        workspaces = (
            backend_client.api.listOrganizations(**DEFAULT_REQUEST_KWARGS)
            .response()
            .result
        )
        workspace_name_to_id = {f"{f.name}": f.id for f in workspaces}
    except HTTPNotFound:
        raise WorkspaceNotFound(workspace=workspace)

    if workspace not in workspace_name_to_id:
        raise WorkspaceNotFound(workspace=workspace)

    params = {
        "projectToCreate": {
            "name": name,
            "description": description,
            "projectKey": key,
            "organizationId": workspace_name_to_id[workspace],
            "visibility": ProjectVisibilityDTO.from_str(visibility).value,
        },
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        response = backend_client.api.createProject(**params).response()
        return normalize_project_name(
            name=response.result.name, workspace=response.result.organizationName
        )
    except HTTPBadRequest as e:
        validation_errors = parse_validation_errors(error=e)
        if "ERR_NOT_UNIQUE" in validation_errors:
            raise ProjectAlreadyExists(name=project_identifier) from e
        raise BadRequestException(validation_errors=validation_errors)
    except HTTPUnprocessableEntity as e:
        raise ProjectsLimitReached() from e


@with_api_exceptions_handler
def delete_project(
    name: str, workspace: Optional[str] = None, api_token: Optional[str] = None
):
    """Deletes a project from your Neptune workspace.
    Args:
        name(str): The name of the project in Neptune in the format 'WORKSPACE/PROJECT'.
            If workspace argument was set, it should only contain 'PROJECT' instead of 'WORKSPACE/PROJECT'.
        workspace(str, optional): Name of your Neptune workspace.
            If you specify it, change the format of the name argument to 'PROJECT' instead of 'WORKSPACE/PROJECT'.
            If 'None' it will be parsed from the name argument.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Examples:
        >>> from neptune import management
        >>> management.delete_project(name="awesome-team/amazing-project")
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    """
    verify_type("name", name, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {"projectIdentifier": project_identifier, **DEFAULT_REQUEST_KWARGS}

    try:
        backend_client.api.deleteProject(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnDeletion(name=project_identifier) from e


@with_api_exceptions_handler
def add_project_member(
    name: str,
    username: str,
    role: str,
    workspace: Optional[str] = None,
    api_token: Optional[str] = None,
):
    """Adds member to the Neptune project.
    Args:
        name(str): The name of the project in Neptune in the format 'WORKSPACE/PROJECT'.
            If workspace argument was set, it should only contain 'PROJECT' instead of 'WORKSPACE/PROJECT'.
        username(str): Name of the user you want to add to the project.
        role(str): level of permissions the user should have in a project.
            Can be set to:
             - 'viewer': can only view project content and members
             - 'contributor': can view and edit project content and only view members
             - 'owner': can view and edit project content and members
            For more information, see `user roles in a project docs`_.
        workspace(str, optional): Name of your Neptune workspace.
            If you specify it, change the format of the name argument to 'PROJECT' instead of 'WORKSPACE/PROJECT'.
            If 'None' it will be parsed from the name argument.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Examples:
        >>> from neptune import management
        >>> management.add_project_member(name="awesome-team/amazing-project",
        ...                               username="johny",
        ...                               role="contributor")
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    .. _user roles in a project docs:
       https://docs.neptune.ai/administration/user-management#roles-in-a-project
    """
    verify_type("name", name, str)
    verify_type("username", username, str)
    verify_type("role", role, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {
        "projectIdentifier": project_identifier,
        "member": {
            "userId": username,
            "role": ProjectMemberRoleDTO.from_str(role).value,
        },
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        backend_client.api.addProjectMember(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPConflict as e:
        raise UserAlreadyHasAccess(user=username, project=project_identifier) from e


@with_api_exceptions_handler
def get_project_member_list(
    name: str, workspace: Optional[str] = None, api_token: Optional[str] = None
) -> Dict[str, str]:
    """Get a list of members for a project.
    Args:
        name(str): The name of the project in Neptune in the format 'WORKSPACE/PROJECT'.
            If workspace argument was set it should only contain 'PROJECT' instead of 'WORKSPACE/PROJECT'.
        workspace(str, optional): Name of your Neptune workspace.
            If you specify change the format of the name argument to 'PROJECT' instead of 'WORKSPACE/PROJECT'.
            If 'None' it will be parsed from the name argument.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Returns:
        ``Dict[str, str]``: Dictionary with usernames as keys and ProjectMemberRoles
        ('owner', 'contributor', 'viewer') as values.
    Examples:
        >>> from neptune import management
        >>> management.get_project_member_list(name="awesome-team/amazing-project")
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    """
    verify_type("name", name, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {"projectIdentifier": project_identifier, **DEFAULT_REQUEST_KWARGS}

    try:
        result = backend_client.api.listProjectMembers(**params).response().result
        return {
            f"{m.registeredMemberInfo.username}": ProjectMemberRoleDTO.to_domain(m.role)
            for m in result
        }
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e


@with_api_exceptions_handler
def remove_project_member(
    name: str,
    username: str,
    workspace: Optional[str] = None,
    api_token: Optional[str] = None,
):
    """Removes member from the Neptune project.
    Args:
        name(str): The name of the project in Neptune in the format 'WORKSPACE/PROJECT'.
            If workspace argument was set, it should only contain 'PROJECT' instead of 'WORKSPACE/PROJECT'.
        username(str): name of the user you want to remove from the project.
        workspace(str, optional): Name of your Neptune workspace.
            If you specify change the format of the name argument to 'PROJECT' instead of 'WORKSPACE/PROJECT'.
            If 'None' it will be parsed from the name argument.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Examples:
        >>> from neptune import management
        >>> management.remove_project_member(name="awesome-team/amazing-project",
        ...                                  username="johny")
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    """
    verify_type("name", name, str)
    verify_type("username", username, str)
    verify_type("workspace", workspace, (str, type(None)))
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {
        "projectIdentifier": project_identifier,
        "userId": username,
        **DEFAULT_REQUEST_KWARGS,
    }

    try:
        backend_client.api.deleteProjectMember(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPUnprocessableEntity as e:
        raise UserNotExistsOrWithoutAccess(
            user=username, project=project_identifier
        ) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnMemberRemoval(
            user=username, project=project_identifier
        ) from e


@with_api_exceptions_handler
def get_workspace_member_list(
    name: str, api_token: Optional[str] = None
) -> Dict[str, str]:
    """Get a list of members of a workspace.
    Args:
        name(str, optional): Name of your Neptune workspace.
        api_token(str, optional): User’s API token. Defaults to `None`.
            If `None`, the value of `NEPTUNE_API_TOKEN` environment variable will be taken.
            .. note::
                It is strongly recommended to use `NEPTUNE_API_TOKEN` environment variable rather than placing your
                API token in plain text in your source code.
    Returns:
        ``Dict[str, str]``: Dictionary with usernames as keys and `WorkspaceMemberRole` ('member', 'admin') as values.
    Examples:
        >>> from neptune import management
        >>> management.get_workspace_member_list(name="awesome-team")
    You may also want to check `management API reference`_.
    .. _management API reference:
       https://docs.neptune.ai/api-reference/management
    """
    verify_type("name", name, str)
    verify_type("api_token", api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)

    params = {"organizationIdentifier": name, **DEFAULT_REQUEST_KWARGS}

    try:
        result = backend_client.api.listOrganizationMembers(**params).response().result
        return {
            f"{m.registeredMemberInfo.username}": WorkspaceMemberRoleDTO.to_domain(
                m.role
            )
            for m in result
        }
    except HTTPNotFound as e:
        raise WorkspaceNotFound(workspace=name) from e



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

from faker import Faker
import boto3
import pytest

from neptune.management.internal.utils import normalize_project_name
from neptune.management import create_project, add_project_member
import neptune.new as neptune

from e2e_tests.utils import a_project_name, Environment

fake = Faker()


@pytest.fixture(scope="session")
def environment():
    workspace = os.getenv("WORKSPACE_NAME")
    admin_token = os.getenv("ADMIN_NEPTUNE_API_TOKEN")
    user = os.getenv("USER_USERNAME")

    project_name, project_key = a_project_name(project_slug=fake.slug())
    project_identifier = normalize_project_name(name=project_name, workspace=workspace)
    created_project_identifier = create_project(
        name=project_name,
        key=project_key,
        visibility="priv",
        workspace=workspace,
        api_token=admin_token,
    )

    add_project_member(
        name=created_project_identifier,
        username=user,
        # pylint: disable=no-member
        role="contributor",
        api_token=admin_token,
    )

    yield Environment(
        workspace=workspace,
        project=project_identifier,
        user_token=os.getenv("NEPTUNE_API_TOKEN"),
        admin_token=admin_token,
        admin=os.getenv("ADMIN_USERNAME"),
        user=user,
    )


@pytest.fixture(scope="session")
def container(request, environment):
    if request.param == "project":
        project = neptune.init_project(name=environment.project)
        yield project
        project.stop()

    if request.param == "run":
        exp = neptune.init_run(project=environment.project)
        yield exp
        exp.stop()


@pytest.fixture(scope="session")
def bucket(environment):
    bucket_name = os.environ.get("BUCKET_NAME")

    s3_client = boto3.resource("s3")
    s3_bucket = s3_client.Bucket(bucket_name)

    yield bucket_name, s3_client

    s3_bucket.objects.filter(Prefix=environment.project).delete()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import re
from pathlib import Path

import neptune.new as neptune
from click.testing import CliRunner
from faker import Faker
from neptune.new.sync import sync

from e2e_tests.base import BaseE2ETest
from e2e_tests.utils import DISABLE_SYSLOG_KWARGS, tmp_context

fake = Faker()
runner = CliRunner()


class TestSync(BaseE2ETest):
    SYNCHRONIZED_SYSID_RE = r"\w+/[\w-]+/([\w-]+)"

    def test_sync_run(self, environment):
        custom_run_id = "-".join((fake.word() for _ in range(3)))

        with tmp_context() as tmp:
            # with test values
            key = self.gen_key()
            original_value = fake.word()
            updated_value = fake.word()

            # init run
            run = neptune.init(
                custom_run_id=custom_run_id,
                project=environment.project,
                **DISABLE_SYSLOG_KWARGS,
            )

            def get_next_run():
                return neptune.init(
                    custom_run_id=custom_run_id,
                    project=environment.project,
                    **DISABLE_SYSLOG_KWARGS,
                )

            self._test_sync(
                exp=run,
                get_next_exp=get_next_run,
                path=tmp,
                key=key,
                original_value=original_value,
                updated_value=updated_value,
            )

    def test_sync_project(self, environment):
        with tmp_context() as tmp:
            # with test values
            key = f"{self.gen_key()}-" + "-".join((fake.word() for _ in range(3)))
            original_value = fake.word()
            updated_value = fake.word()

            # init run
            project = neptune.init_project(name=environment.project)

            def get_next_project():
                return neptune.init_project(name=environment.project)

            self._test_sync(
                exp=project,
                get_next_exp=get_next_project,
                path=tmp,
                key=key,
                original_value=original_value,
                updated_value=updated_value,
            )

    @staticmethod
    def _test_sync(exp, get_next_exp, path, key, original_value, updated_value):
        # assign original value
        exp[key] = original_value
        exp.sync()

        # stop run
        exp.stop()

        # pylint: disable=protected-access
        queue_dir = list(Path(f"./.neptune/async/{exp._id}/").glob("exec-*"))[0]
        with open(
            queue_dir / "last_put_version", encoding="utf-8"
        ) as last_put_version_f:
            last_put_version = int(last_put_version_f.read())
        with open(queue_dir / "data-1.log", "a", encoding="utf-8") as queue_f:
            queue_f.write(
                json.dumps(
                    {
                        "obj": {
                            "type": "AssignString",
                            "path": key.split("/"),
                            "value": updated_value,
                        },
                        "version": last_put_version + 1,
                    }
                )
            )
            queue_f.write(
                json.dumps(
                    {
                        "obj": {
                            "type": "CopyAttribute",
                            "path": ["copy"] + key.split("/"),
                            "container_id": exp._id,
                            "container_type": exp.container_type.value,
                            "source_path": key.split("/"),
                            "source_attr_name": "String",
                        },
                        "version": last_put_version + 2,
                    }
                )
            )
        with open(
            queue_dir / "last_put_version", "w", encoding="utf-8"
        ) as last_put_version_f:
            last_put_version_f.write(str(last_put_version + 2))

        # other exp should see only original value from server
        exp2 = get_next_exp()
        assert exp2[key].fetch() == original_value

        # run neptune sync
        result = runner.invoke(sync, ["--path", path])
        assert result.exit_code == 0

        # other exp should see updated value from server
        exp3 = get_next_exp()
        assert exp3[key].fetch() == updated_value
        assert exp3["copy/" + key].fetch() == updated_value

    def test_offline_sync(self, environment):
        with tmp_context() as tmp:
            # create run in offline mode
            run = neptune.init(
                mode="offline",
                project=environment.project,
                **DISABLE_SYSLOG_KWARGS,
            )
            # assign some values
            key = self.gen_key()
            val = fake.word()
            run[key] = val

            # and stop it
            run.stop()

            # run asynchronously
            result = runner.invoke(sync, ["--path", tmp, "-p", environment.project])
            assert result.exit_code == 0

            # offline mode doesn't support custom_run_id, we'll have to parse sync output to determine short_id
            sys_id_found = re.search(self.SYNCHRONIZED_SYSID_RE, result.stdout)
            assert len(sys_id_found.groups()) == 1
            sys_id = sys_id_found.group(1)

            run2 = neptune.init(run=sys_id, project=environment.project)
            assert run2[key].fetch() == val



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=redefined-outer-name

import os
import random
import string
import threading
import uuid
from random import randint

import pytest

import neptune.new.sync
from neptune.new.constants import OFFLINE_DIRECTORY
from neptune.new.exceptions import ProjectNotFound
from neptune.new.internal.backends.api_model import Project
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.containers.disk_queue import DiskQueue
from neptune.new.internal.operation import Operation
from neptune.new.internal.utils.sync_offset_file import SyncOffsetFile
from neptune.new.sync import (
    ApiRun,
    get_project,
    get_qualified_name,
    sync_all_runs,
    sync_selected_runs,
    synchronization_status,
)


def a_run():
    return ApiRun(
        str(uuid.uuid4()), "RUN-{}".format(randint(42, 12342)), "org", "proj", False
    )


def a_project():
    return ApiRun(
        str(uuid.uuid4()),
        "".join((random.choice(string.ascii_letters).upper() for _ in range(3))),
        "org",
        "proj",
        False,
    )


def generate_get_run_impl(registered_experiments):
    def get_run_impl(run_id):
        """This function will return run as well as projects. Will be cleaned in ModelRegistry"""
        for exp in registered_experiments:
            if run_id in (str(exp.id), get_qualified_name(exp)):
                return exp

    return get_run_impl


def prepare_projects(path):
    unsync_project = a_project()
    sync_project = a_project()
    registered_projects = (unsync_project, sync_project)

    execution_id = "exec-0"

    for project in registered_projects:
        project_path = path / "async" / str(project.id) / execution_id
        project_path.mkdir(parents=True)
        queue = DiskQueue(
            project_path,
            lambda x: x,
            lambda x: x,
            threading.RLock(),
            ContainerType.PROJECT,
        )
        queue.put("op-proj-0")
        queue.put("op-proj-1")

    SyncOffsetFile(
        path / "async" / str(unsync_project.id) / execution_id / "last_ack_version"
    ).write(1)
    SyncOffsetFile(
        path / "async" / str(unsync_project.id) / execution_id / "last_put_version"
    ).write(2)

    SyncOffsetFile(
        path / "async" / str(sync_project.id) / execution_id / "last_ack_version"
    ).write(2)
    SyncOffsetFile(
        path / "async" / str(sync_project.id) / execution_id / "last_put_version"
    ).write(2)

    return unsync_project, sync_project, generate_get_run_impl(registered_projects)


def prepare_runs(path):
    unsync_exp = a_run()
    sync_exp = a_run()
    registered_runs = (unsync_exp, sync_exp)

    execution_id = "exec-0"

    for exp in registered_runs:
        exp_path = path / "async" / str(exp.id) / execution_id
        exp_path.mkdir(parents=True)
        queue = DiskQueue(
            exp_path, lambda x: x, lambda x: x, threading.RLock(), ContainerType.RUN
        )
        queue.put("op-0")
        queue.put("op-1")

    SyncOffsetFile(
        path / "async" / str(unsync_exp.id) / execution_id / "last_ack_version"
    ).write(1)
    SyncOffsetFile(
        path / "async" / str(unsync_exp.id) / execution_id / "last_put_version"
    ).write(2)

    SyncOffsetFile(
        path / "async" / str(sync_exp.id) / execution_id / "last_ack_version"
    ).write(2)
    SyncOffsetFile(
        path / "async" / str(sync_exp.id) / execution_id / "last_put_version"
    ).write(2)

    return unsync_exp, sync_exp, generate_get_run_impl(registered_runs)


def prepare_offline_run(path):
    offline_exp_uuid = str(uuid.uuid4())
    offline_exp_path = path / OFFLINE_DIRECTORY / offline_exp_uuid
    offline_exp_path.mkdir(parents=True)

    queue = DiskQueue(
        offline_exp_path, lambda x: x, lambda x: x, threading.RLock(), ContainerType.RUN
    )
    queue.put("op-0")
    queue.put("op-1")
    SyncOffsetFile(
        path / OFFLINE_DIRECTORY / offline_exp_uuid / "last_put_version"
    ).write(2)

    return offline_exp_uuid


def test_list_projects(tmp_path, mocker, capsys):
    """TODO: we're mentioning projects as runs, will be improved with ModelRegistry"""
    # given
    unsync_proj, sync_proj, get_exp_impl = prepare_projects(tmp_path)
    offline_exp_uuid = prepare_offline_run(tmp_path)

    # and
    mocker.patch.object(neptune.new.sync, "get_run", get_exp_impl)
    mocker.patch.object(Operation, "from_dict")

    # when
    synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        "Synchronized runs:\n- {}".format(get_qualified_name(sync_proj)) in captured.out
    )
    assert (
        "Unsynchronized runs:\n- {}".format(get_qualified_name(unsync_proj))
        in captured.out
    )
    assert (
        "Unsynchronized offline runs:\n- offline/{}".format(offline_exp_uuid)
        in captured.out
    )


def test_list_runs(tmp_path, mocker, capsys):
    # given
    unsync_exp, sync_exp, get_run_impl = prepare_runs(tmp_path)
    offline_exp_uuid = prepare_offline_run(tmp_path)

    # and
    mocker.patch.object(neptune.new.sync, "get_run", get_run_impl)
    mocker.patch.object(Operation, "from_dict")

    # when
    synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        "Synchronized runs:\n- {}".format(get_qualified_name(sync_exp)) in captured.out
    )
    assert (
        "Unsynchronized runs:\n- {}".format(get_qualified_name(unsync_exp))
        in captured.out
    )
    assert (
        "Unsynchronized offline runs:\n- offline/{}".format(offline_exp_uuid)
        in captured.out
    )


def test_list_runs_when_no_run(tmp_path, capsys):
    (tmp_path / "async").mkdir()
    # when
    with pytest.raises(SystemExit):
        synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert "There are no Neptune runs" in captured.out


def test_sync_all_runs(tmp_path, mocker, capsys):
    # given
    unsync_proj, sync_proj, _ = prepare_projects(tmp_path)
    unsync_exp, sync_exp, _ = prepare_runs(tmp_path)
    get_run_impl = generate_get_run_impl((unsync_proj, sync_proj, unsync_exp, sync_exp))
    offline_exp_uuid = prepare_offline_run(tmp_path)
    registered_offline_run = a_run()

    # and
    mocker.patch.object(neptune.new.sync, "get_run", get_run_impl)
    mocker.patch.object(neptune.new.sync, "backend")
    mocker.patch.object(neptune.new.sync.backend, "execute_operations")
    mocker.patch.object(
        neptune.new.sync.backend,
        "get_project",
        lambda _: Project(str(uuid.uuid4()), "project", "workspace"),
    )
    mocker.patch.object(
        neptune.new.sync,
        "register_offline_run",
        lambda project, container_type: (registered_offline_run, True),
    )
    mocker.patch.object(Operation, "from_dict", lambda x: x)
    neptune.new.sync.backend.execute_operations.return_value = (1, [])

    # when
    sync_all_runs(tmp_path, "foo")

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        "Offline run {} registered as {}".format(
            offline_exp_uuid, get_qualified_name(registered_offline_run)
        )
    ) in captured.out
    assert "Synchronising {}".format(get_qualified_name(unsync_exp)) in captured.out
    assert "Synchronising {}".format(get_qualified_name(unsync_proj)) in captured.out
    assert (
        "Synchronization of run {} completed.".format(get_qualified_name(unsync_exp))
        in captured.out
    )
    assert (
        "Synchronization of project {} completed.".format(
            get_qualified_name(unsync_proj)
        )
        in captured.out
    )
    assert "Synchronising {}".format(get_qualified_name(sync_exp)) not in captured.out
    assert "Synchronising {}".format(get_qualified_name(sync_proj)) not in captured.out

    # and
    # pylint: disable=no-member
    neptune.new.sync.backend.execute_operations.has_calls(
        [
            mocker.call(unsync_exp.id, ContainerType.RUN, ["op-1"]),
            mocker.call(registered_offline_run.id, ContainerType.RUN, ["op-1"]),
            mocker.call(unsync_proj.id, ContainerType.PROJECT, ["op-proj-1"]),
        ],
        any_order=True,
    )


def test_sync_selected_runs(tmp_path, mocker, capsys):
    # given
    unsync_exp, sync_exp, get_run_impl = prepare_runs(tmp_path)
    offline_exp_uuid = prepare_offline_run(tmp_path)
    registered_offline_exp = a_run()

    def get_run_impl_(run_id: str):
        if run_id in (
            str(registered_offline_exp.id),
            get_qualified_name(registered_offline_exp),
        ):
            return registered_offline_exp
        else:
            return get_run_impl(run_id)

    # and
    mocker.patch.object(neptune.new.sync, "get_run", get_run_impl_)
    mocker.patch.object(neptune.new.sync, "backend")
    mocker.patch.object(neptune.new.sync.backend, "execute_operations")
    mocker.patch.object(
        neptune.new.sync.backend,
        "get_project",
        lambda _: Project(str(uuid.uuid4()), "project", "workspace"),
    )
    mocker.patch.object(
        neptune.new.sync,
        "register_offline_run",
        lambda project, container_type: (registered_offline_exp, True),
    )
    mocker.patch.object(Operation, "from_dict", lambda x: x)
    neptune.new.sync.backend.execute_operations.return_value = (2, [])

    # when
    sync_selected_runs(
        tmp_path,
        "some-name",
        [get_qualified_name(sync_exp), "offline/" + offline_exp_uuid],
    )

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert "Synchronising {}".format(get_qualified_name(sync_exp)) in captured.out
    assert (
        "Synchronization of run {} completed.".format(get_qualified_name(sync_exp))
        in captured.out
    )
    assert (
        "Synchronising {}".format(get_qualified_name(registered_offline_exp))
        in captured.out
    )
    assert (
        "Synchronization of run {} completed.".format(
            get_qualified_name(registered_offline_exp)
        )
        in captured.out
    )
    assert "Synchronising {}".format(get_qualified_name(unsync_exp)) not in captured.out

    # and
    # pylint: disable=no-member
    neptune.new.sync.backend.execute_operations.assert_called_with(
        registered_offline_exp.id, ContainerType.RUN, operations=["op-0", "op-1"]
    )


def test_get_project_no_name_set(mocker):
    # given
    mocker.patch.object(os, "getenv")
    os.getenv.return_value = None

    # expect
    assert get_project(None) is None


def test_get_project_project_not_found(mocker):
    # given
    mocker.patch.object(neptune.new.sync, "backend")
    mocker.patch.object(neptune.new.sync.backend, "get_project")
    neptune.new.sync.backend.get_project.side_effect = ProjectNotFound("foo")

    # expect
    assert get_project("foo") is None


def test_sync_non_existent_run(tmp_path, mocker, capsys):
    # given
    mocker.patch.object(neptune.new.sync, "get_project")
    mocker.patch.object(neptune.new.sync, "get_run")
    neptune.new.sync.get_run.return_value = a_run()

    # when
    sync_selected_runs(tmp_path, "foo", ["bar"])

    # then
    captured = capsys.readouterr()
    assert "Warning: Run 'bar' does not exist in location" in captured.err



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import threading
from typing import Union, Optional, Dict, Any

import click

from neptune.new.attribute_container import AttributeContainer
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.operation_processors.operation_processor import (
    OperationProcessor,
)
from neptune.new.internal.state import ContainerState

# backwards compatibility
# pylint: disable=unused-import
from neptune.new.attributes.attribute import Attribute
from neptune.new.attributes.namespace import (
    NamespaceBuilder,
    Namespace as NamespaceAttr,
)
from neptune.new.exceptions import (
    MetadataInconsistency,
    InactiveRunException,
    NeptunePossibleLegacyUsageException,
)
from neptune.new.handler import Handler
from neptune.new.internal.backends.api_model import AttributeType
from neptune.new.internal.operation import DeleteAttribute
from neptune.new.internal.run_structure import ContainerStructure as RunStructure
from neptune.new.internal.utils import (
    is_bool,
    is_float,
    is_float_like,
    is_int,
    is_string,
    is_string_like,
    verify_type,
    is_dict_like,
)
from neptune.new.internal.utils.paths import parse_path
from neptune.new.internal.value_to_attribute_visitor import ValueToAttributeVisitor
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.namespace import Namespace
from neptune.new.types.value import Value

RunState = ContainerState


class Run(AttributeContainer):
    """A Run in Neptune is a representation of all metadata that you log to Neptune.

    Beginning when you start a tracked run with `neptune.init()` and ending when the script finishes
    or when you explicitly stop the experiment with `.stop()`.

    You can log many ML metadata types, including:
        * metrics
        * losses
        * model weights
        * images
        * interactive charts
        * predictions
        * and much more

    Examples:
        >>> import neptune.new as neptune

        >>> # Create new experiment
        ... run = neptune.init('my_workspace/my_project')

        >>> # Log parameters
        ... params = {'max_epochs': 10, 'optimizer': 'Adam'}
        ... run['parameters'] = params

        >>> # Log metadata
        ... run['train/metric_name'].log()
        >>> run['predictions'].log(image)
        >>> run['model'].upload(path_to_model)

        >>> # Log whatever else you want
        ... pass

        >>> # Stop tracking and clean up
        ... run.stop()

    You may also want to check `Run docs page`_.

    .. _Run docs page:
       https://docs.neptune.ai/api-reference/run
    """

    last_run = None  # "static" instance of recently created Run

    container_type = ContainerType.RUN

    LEGACY_METHODS = (
        "create_experiment",
        "send_metric",
        "log_metric",
        "send_text",
        "log_text",
        "send_image",
        "log_image",
        "send_artifact",
        "log_artifact",
        "delete_artifacts",
        "download_artifact",
        "download_sources",
        "download_artifacts",
        "reset_log",
        "get_parameters",
        "get_properties",
        "set_property",
        "remove_property",
        "get_hardware_utilization",
        "get_numeric_channels_values",
    )

    def __init__(
        self,
        _id: str,
        backend: NeptuneBackend,
        op_processor: OperationProcessor,
        background_job: BackgroundJob,
        lock: threading.RLock,
        workspace: str,
        project_name: str,
        short_id: str,
        project_id: str,
        monitoring_namespace: str = "monitoring",
    ):
        super().__init__(
            _id,
            backend,
            op_processor,
            background_job,
            lock,
            project_id,
            project_name,
            workspace,
        )
        self._short_id = short_id
        self.monitoring_namespace = monitoring_namespace

        Run.last_run = self

    @property
    def _label(self) -> str:
        return self._short_id

    @property
    def _docs_url_stop(self) -> str:
        return "https://docs.neptune.ai/api-reference/run#.stop"

    def get_run_url(self) -> str:
        """Returns the URL the run can be accessed with in the browser"""
        return self._backend.get_run_url(
            self._id, self._workspace, self._project_name, self._short_id
        )

    def _startup(self, debug_mode):
        if not debug_mode:
            click.echo(self.get_run_url())
        super()._startup(debug_mode)

    def assign(self, value, wait: bool = False) -> None:
        """Assign values to multiple fields from a dictionary.
        You can use this method to quickly log all run's parameters.

        Args:
            value (dict): A dictionary with values to assign, where keys become the paths of the fields.
                The dictionary can be nested - in such case the path will be a combination of all keys.
            wait (bool, optional): If `True` the client will first wait to send all tracked metadata to the server.
                This makes the call synchronous. Defaults to `False`.

        Examples:
            >>> import neptune.new as neptune
            >>> run = neptune.init_run()

            >>> # Assign multiple fields from a dictionary
            ... params = {"max_epochs": 10, "optimizer": "Adam"}
            >>> run["parameters"] = params

            >>> # You can always log explicitly parameters one by one
            ... run["parameters/max_epochs"] = 10
            >>> run["parameters/optimizer"] = "Adam"

            >>> # Dictionaries can be nested
            ... params = {"train": {"max_epochs": 10}}
            >>> run["parameters"] = params
            >>> # This will log 10 under path "parameters/train/max_epochs"

        You may also want to check `assign docs page`_.

        .. _assign docs page:
            https://docs.neptune.ai/api-reference/run#.assign
        """
        return AttributeContainer.assign(self, value=value, wait=wait)

    def fetch(self) -> dict:
        """Fetch values of all non-File Atom fields as a dictionary.
        The result will preserve the hierarchical structure of the run's metadata, but will contain only non-File Atom
        fields.
        You can use this method to quickly retrieve previous run's parameters.

        Returns:
            `dict` containing all non-File Atom fields values.

        Examples:
            >>> import neptune.new as neptune
            >>> resumed_run = neptune.init(run="HEL-3")
            >>> params = resumed_run['model/parameters'].fetch()

            >>> run_data = resumed_run.fetch()

            >>> print(run_data)
            >>> # this will print out all Atom attributes stored in run as a dict

        You may also want to check `fetch docs page`_.

        .. _fetch docs page:
            https://docs.neptune.ai/api-reference/run#.fetch
        """
        return AttributeContainer.fetch(self)

    def stop(self, seconds: Optional[Union[float, int]] = None) -> None:
        """Stops the tracked run and kills the synchronization thread.

        `.stop()` will be automatically called when a script that created the run finishes or on the destruction
        of Neptune context.

        When using Neptune with Jupyter notebooks it's a good practice to stop the tracked run manually as it
        will be stopped automatically only when the Jupyter kernel stops.

        Args:
            seconds (int or float, optional): Seconds to wait for all tracking calls to finish
                before stopping the tracked run.
                If `None` will wait for all tracking calls to finish. Defaults to `True`.

        Examples:
            If you are creating tracked runs from the script you don't need to call `.stop()`:

            >>> import neptune.new as neptune
            >>> run = neptune.init()

            >>> # Your training or monitoring code
            ... pass
            ... # If you are executing Python script .stop()
            ... # is automatically called at the end for every run

            If you are performing multiple training jobs from one script one after the other it is a good practice
            to `.stop()` the finished tracked runs as every open run keeps an open connection with Neptune,
            monitors hardware usage, etc. You can also use Context Managers - Neptune will automatically call `.stop()`
            on the destruction of Run context:

            >>> import neptune.new as neptune

            >>> # If you are running consecutive training jobs from the same script
            ... # stop the tracked runs manually at the end of single training job
            ... for config in configs:
            ...   run = neptune.init()
            ...   # Your training or monitoring code
            ...   pass
            ...   run.stop()

            >>> # You can also use with statement and context manager
            ... for config in configs:
            ...   with neptune.init() as run:
            ...     # Your training or monitoring code
            ...     pass
            ...     # .stop() is automatically called
            ...     # when code execution exits the with statement

        .. warning::
            If you are using Jupyter notebooks for creating your runs you need to manually invoke `.stop()` once the
            training and evaluation is done.

        You may also want to check `stop docs page`_.

        .. _stop docs page:
            https://docs.neptune.ai/api-reference/run#.stop
        """
        return AttributeContainer.stop(self, seconds=seconds)

    def get_structure(self) -> Dict[str, Any]:
        """Returns a run's metadata structure in form of a dictionary.

        This method can be used to traverse the run's metadata structure programmatically
        when using Neptune in automated workflows.

        .. danger::
            The returned object is a deep copy of an internal run's structure.

        Returns:
            ``dict``: with the run's metadata structure.

        """
        return AttributeContainer.get_structure(self)

    def print_structure(self) -> None:
        """Pretty prints the structure of the run's metadata.

        Paths are ordered lexicographically and the whole structure is neatly colored.
        """
        return AttributeContainer.print_structure(self)

    def pop(self, path: str, wait: bool = False) -> None:
        """Removes the field stored under the path completely and all data associated with it.

        Args:
            path (str): Path of the field to be removed.
            wait (bool, optional): If `True` the client will first wait to send all tracked metadata to the server.
                This makes the call synchronous. Defaults to `True`.

        Examples:
            >>> import neptune.new as neptune
            >>> run = neptune.init()

            >>> run['parameters/learninggg_rata'] = 0.3

            >>> # Delete a field along with it's data
            ... run.pop('parameters/learninggg_rata')

            >>> run['parameters/learning_rate'] = 0.3

            >>> # Training finished
            ... run['trained_model'].upload('model.pt')
            >>> # 'model_checkpoint' is a File field
            ... run.pop('model_checkpoint')

        You may also want to check `pop docs page`_.

        .. _pop docs page:
           https://docs.neptune.ai/api-reference/run#.pop
        """
        return AttributeContainer.pop(self, path=path, wait=wait)

    def wait(self, disk_only=False) -> None:
        """Wait for all the tracking calls to finish.

        Args:
            disk_only (bool, optional, default is False): If `True` the process will only wait for data to be saved
                locally from memory, but will not wait for them to reach Neptune servers.
                Defaults to `False`.

        You may also want to check `wait docs page`_.

        .. _wait docs page:
            https://docs.neptune.ai/api-reference/run#.wait
        """
        return AttributeContainer.wait(self, disk_only=disk_only)

    def sync(self, wait: bool = True) -> None:
        """Synchronizes local representation of the run with Neptune servers.

        Args:
            wait (bool, optional, default is True): If `True` the process will only wait for data to be saved
                locally from memory, but will not wait for them to reach Neptune servers.
                Defaults to `True`.

        Examples:
            >>> import neptune.new as neptune

            >>> # Connect to a run from Worker #3
            ... worker_id = 3
            >>> run = neptune.init(run='DIST-43', monitoring_namespace='monitoring/{}'.format(worker_id))

            >>> # Try to access logs that were created in meantime by Worker #2
            ... worker_2_status = run['status/2'].fetch() # Error if this field was created after this script starts

            >>> run.sync() # Synchronizes local representation with Neptune servers.
            >>> worker_2_status = run['status/2'].fetch() # No error

        You may also want to check `sync docs page`_.

        .. _sync docs page:
            https://docs.neptune.ai/api-reference/run#.sync
        """
        return AttributeContainer.sync(self, wait=wait)



import os
import uuid
from zipfile import ZipFile

import pytest

from e2e_tests.base import BaseE2ETest
from e2e_tests.standard.test_base import fake
from e2e_tests.utils import tmp_context
from neptune.new.attribute_container import AttributeContainer
from neptune.new.internal.backends.api_model import OptionalFeatures, MultipartConfig
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend


class TestUpload(BaseE2ETest):
    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    def test_using_new_api(self, container: AttributeContainer):
        # pylint: disable=protected-access
        assert isinstance(container._backend, HostedNeptuneBackend)
        assert container._backend._client_config.has_feature(
            OptionalFeatures.MULTIPART_UPLOAD
        )
        assert isinstance(
            container._backend._client_config.multipart_config, MultipartConfig
        )

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    @pytest.mark.parametrize(
        "file_size",
        [
            pytest.param(10 * 2 ** 20, id="big"),  # 10 MB, multipart
            pytest.param(100 * 2 ** 10, id="small"),  # 100 kB, single upload
        ],
    )
    def test_single_file(self, container: AttributeContainer, file_size: int):
        key = self.gen_key()
        filename = fake.file_name()
        downloaded_filename = fake.file_name()

        with tmp_context():
            # create 10MB file
            with open(filename, "wb") as file:
                file.write(b"\0" * file_size)
            container[key].upload(filename)

            container.sync()
            container[key].download(downloaded_filename)

            assert os.path.getsize(downloaded_filename) == file_size
            with open(downloaded_filename, "rb") as file:
                content = file.read()
                assert len(content) == file_size
                assert content == b"\0" * file_size

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    def test_fileset(self, container: AttributeContainer):
        key = self.gen_key()
        large_filesize = 10 * 2 ** 20
        large_filename = fake.file_name()
        small_files = [
            (f"{uuid.uuid4()}.{fake.file_extension()}", fake.sentence().encode("utf-8"))
            for _ in range(100)
        ]

        with tmp_context():
            # create single large file (multipart) and a lot of very small files
            with open(large_filename, "wb") as file:
                file.write(b"\0" * large_filesize)
            for filename, contents in small_files:
                with open(filename, "wb") as file:
                    file.write(contents)

            small_filenames = [filename for filename, _ in small_files]
            # make sure there are no duplicates
            assert len({large_filename, *small_filenames}) == len(small_files) + 1

            # when one file as fileset uploaded
            container[key].upload_files([large_filename])

            # then check if will be downloaded
            container.sync()
            container[key].download("downloaded1.zip")

            with ZipFile("downloaded1.zip") as zipped:
                assert set(zipped.namelist()) == {large_filename, "/"}
                with zipped.open(large_filename, "r") as file:
                    content = file.read()
                    assert len(content) == large_filesize
                    assert content == b"\0" * large_filesize

            # when small files as fileset uploaded
            container[key].upload_files(small_filenames)

            # then check if everything will be downloaded
            container.sync()
            container[key].download("downloaded2.zip")

            with ZipFile("downloaded2.zip") as zipped:
                assert set(zipped.namelist()) == {large_filename, "/", *small_filenames}
                with zipped.open(large_filename, "r") as file:
                    content = file.read()
                    assert len(content) == large_filesize
                    assert content == b"\0" * large_filesize
                for filename, expected_content in small_files:
                    with zipped.open(filename, "r") as file:
                        content = file.read()
                        assert len(content) == len(expected_content)
                        assert content == expected_content

            # when first file is removed
            container[key].delete_files([large_filename])

            # then check if the rest will be downloaded
            container.sync()
            container[key].download("downloaded3.zip")

            with ZipFile("downloaded3.zip") as zipped:
                assert set(zipped.namelist()) == {"/", *small_filenames}
                for filename, expected_content in small_files:
                    with zipped.open(filename, "r") as file:
                        content = file.read()
                        assert len(content) == len(expected_content)
                        assert content == expected_content



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import random
import pytest
from faker import Faker

import neptune.new as neptune
from neptune.new.run import Run
from neptune.new.project import Project

from e2e_tests.base import BaseE2ETest

fake = Faker()


class TestCopying(BaseE2ETest):
    @pytest.mark.parametrize("container", ["run", "project"], indirect=True)
    @pytest.mark.parametrize(
        "value", [random.randint(0, 100), random.random(), fake.boolean(), fake.word()]
    )
    def test_copy_project_to_container(self, container: Run, value, environment):
        project = neptune.init_project(name=environment.project)

        src, destination, destination2 = self.gen_key(), self.gen_key(), self.gen_key()

        project[src] = value
        project.sync()

        container[destination] = project[src]
        container[destination2] = container[destination]
        container.sync()

        assert project[src].fetch() == value
        assert container[destination].fetch() == value
        assert container[destination2].fetch() == value

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    @pytest.mark.parametrize(
        "value", [random.randint(0, 100), random.random(), fake.boolean(), fake.word()]
    )
    def test_copy_run_to_container(self, container: Project, value, environment):
        run = neptune.init_run(project=environment.project)
        src, destination, destination2 = self.gen_key(), self.gen_key(), self.gen_key()

        container[src] = value
        container.sync()

        run[destination] = container[src]
        run[destination2] = run[destination]
        run.sync()

        assert container[src].fetch() == value
        assert run[destination].fetch() == value
        assert run[destination2].fetch() == value



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

import pytest

import neptune.new as neptune
import numpy as np
import torch
import torch.nn.functional as F
from pytorch_lightning.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from sklearn.metrics import accuracy_score
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import DataLoader, random_split
from torchvision import transforms
from torchvision.datasets import MNIST

import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import NeptuneLogger


class LitModel(pl.LightningModule):
    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Not used, for pylint only"""

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """Not used, for pylint only"""

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Not used, for pylint only"""

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        """Not used, for pylint only"""

    def __init__(self, linear, learning_rate, decay_factor, neptune_logger):
        super().__init__()
        self.linear = linear
        self.learning_rate = learning_rate
        self.decay_factor = decay_factor
        self.train_img_max = 10
        self.train_img = 0
        self.layer_1 = torch.nn.Linear(28 * 28, linear)
        self.layer_2 = torch.nn.Linear(linear, 20)
        self.layer_3 = torch.nn.Linear(20, 10)
        self.neptune_logger = neptune_logger

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = self.layer_1(x)
        x = F.relu(x)
        x = self.layer_2(x)
        x = F.relu(x)
        x = self.layer_3(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        scheduler = LambdaLR(optimizer, lambda epoch: self.decay_factor ** epoch)
        return [optimizer], [scheduler]

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)
        self.log("train/loss", loss, prog_bar=False)
        y_true = y.cpu().detach().numpy()
        y_pred = y_hat.argmax(axis=1).cpu().detach().numpy()
        return {"loss": loss, "y_true": y_true, "y_pred": y_pred}

    def training_epoch_end(self, outputs):
        y_true = np.array([])
        y_pred = np.array([])
        for results_dict in outputs:
            y_true = np.append(y_true, results_dict["y_true"])
            y_pred = np.append(y_pred, results_dict["y_pred"])
        acc = accuracy_score(y_true, y_pred)
        self.log("train/loader_acc", acc)

    def validation_step(self, batch, batch_idx, dataset_idx):
        x, y = batch
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)
        self.log("val/loss", loss, prog_bar=False)
        y_true = y.cpu().detach().numpy()
        y_pred = y_hat.argmax(axis=1).cpu().detach().numpy()
        return {"loss": loss, "y_true": y_true, "y_pred": y_pred}

    def validation_epoch_end(self, outputs):
        for dl_idx in range(2):
            y_true = np.array([])
            y_pred = np.array([])
            for results_dict in outputs[dl_idx]:
                y_true = np.append(y_true, results_dict["y_true"])
                y_pred = np.append(y_pred, results_dict["y_pred"])
            acc = accuracy_score(y_true, y_pred)
            self.log("val/loader_acc", acc)

    def test_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y)
        self.log("test/loss", loss, prog_bar=False)
        y_true = y.cpu().detach().numpy()
        y_pred = y_hat.argmax(axis=1).cpu().detach().numpy()
        for j in np.where(np.not_equal(y_true, y_pred))[0]:
            img = np.squeeze(x[j].cpu().detach().numpy())
            img[img < 0] = 0
            img = img / np.amax(img)
            self.neptune_logger.experiment["model_code/test/misclassified_images"].log(
                neptune.types.File.as_image(img),
                description=f"y_pred={y_pred[j]}, y_true={y_true[j]}",
            )
        return {"loss": loss, "y_true": y_true, "y_pred": y_pred}

    def test_epoch_end(self, outputs):
        y_true = np.array([])
        y_pred = np.array([])
        for results_dict in outputs:
            y_true = np.append(y_true, results_dict["y_true"])
            y_pred = np.append(y_pred, results_dict["y_pred"])
        acc = accuracy_score(y_true, y_pred)
        self.log("test/acc", acc)


class MNISTDataModule(pl.LightningDataModule):
    def predict_dataloader(self) -> EVAL_DATALOADERS:
        # not required for e2e test
        pass

    def __init__(self, batch_size, normalization_vector):
        super().__init__()
        self.batch_size = batch_size
        self.normalization_vector = normalization_vector
        self.mnist_train = None
        self.mnist_val1 = None
        self.mnist_val2 = None
        self.mnist_test = None

    def prepare_data(self):
        MNIST(os.getcwd(), train=True, download=True)
        MNIST(os.getcwd(), train=False, download=True)

    def setup(self, stage=None):
        # transforms
        transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize(
                    self.normalization_vector[0], self.normalization_vector[1]
                ),
            ]
        )
        if stage == "fit":
            mnist_train = MNIST(os.getcwd(), train=True, transform=transform)
            # do not use whole set, to save time spent on training
            self.mnist_train, self.mnist_val1, self.mnist_val2, _ = random_split(
                mnist_train, [5000, 500, 500, 54000]
            )
        if stage == "test":
            self.mnist_test = MNIST(os.getcwd(), train=False, transform=transform)

    def train_dataloader(self):
        mnist_train = DataLoader(
            self.mnist_train, batch_size=self.batch_size, num_workers=4
        )
        return mnist_train

    def val_dataloader(self):
        mnist_val1 = DataLoader(
            self.mnist_val1, batch_size=self.batch_size, num_workers=4
        )
        mnist_val2 = DataLoader(
            self.mnist_val2, batch_size=self.batch_size, num_workers=4
        )
        return [mnist_val1, mnist_val2]

    def test_dataloader(self):
        mnist_test = DataLoader(
            self.mnist_test, batch_size=self.batch_size, num_workers=1
        )
        return mnist_test


@pytest.fixture(scope="session")
def pytorch_run(environment):
    # given
    PARAMS = {
        "max_epochs": 3,
        "save_top_k": 2,
        "learning_rate": 0.005,
        "decay_factor": 0.99,
        "batch_size": 64,
        "linear": 64,
    }
    # and
    run = neptune.init(
        name="Integration pytorch-lightning", project=environment.project
    )
    # and
    model_checkpoint = ModelCheckpoint(
        dirpath="my_model/checkpoints/",
        filename="{epoch:02d}-{val/loss/dataloader_idx_1:.2f}",
        save_weights_only=True,
        save_top_k=PARAMS["save_top_k"],
        save_last=True,
        monitor="val/loss/dataloader_idx_1",
        every_n_epochs=1,
    )
    neptune_logger = NeptuneLogger(run=run, prefix="custom_prefix")
    # and (Subject)
    trainer = pl.Trainer(
        max_epochs=PARAMS["max_epochs"],
        log_every_n_steps=10,
        logger=neptune_logger,
        track_grad_norm=2,
        callbacks=[model_checkpoint],
    )
    model = LitModel(
        linear=PARAMS["linear"],
        learning_rate=PARAMS["learning_rate"],
        decay_factor=PARAMS["decay_factor"],
        neptune_logger=neptune_logger,
    )
    data_module = MNISTDataModule(
        normalization_vector=((0.1307,), (0.3081,)), batch_size=PARAMS["batch_size"]
    )

    # then
    trainer.fit(model, datamodule=data_module)
    trainer.test(model, datamodule=data_module)
    run.sync()

    yield run



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import time
import unittest
import uuid

from mock import ANY, MagicMock, call
import psutil

from neptune.internal.hardware.constants import BYTES_IN_ONE_GB
from neptune.internal.hardware.gauges.cpu import SystemCpuUsageGauge
from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.gauges.memory import SystemMemoryUsageGauge
from neptune.internal.hardware.metrics.metric import Metric, MetricResourceType
from neptune.internal.hardware.metrics.reports.metric_report import (
    MetricReport,
    MetricValue,
)
from neptune.internal.hardware.metrics.service.metric_service_factory import (
    MetricServiceFactory,
)
from neptune.utils import IS_MACOS


class TestMetricServiceIntegration(unittest.TestCase):
    def setUp(self):
        self.backend = MagicMock()
        self.metric_service_factory = MetricServiceFactory(
            backend=self.backend, os_environ=os.environ
        )

    @unittest.skipIf(IS_MACOS, "MacOS behaves strangely")
    def test_create_system_metrics(self):
        # given
        memory_amount_gb = psutil.virtual_memory().total / float(BYTES_IN_ONE_GB)

        # and
        experiment = MagicMock()

        # and
        cpu_metric_id = str(uuid.uuid4())
        ram_metric_id = str(uuid.uuid4())
        self.backend.create_hardware_metric.side_effect = [cpu_metric_id, ram_metric_id]

        # when
        self.metric_service_factory.create(
            gauge_mode=GaugeMode.SYSTEM,
            experiment=experiment,
            reference_timestamp=time.time(),
        )

        # then
        self.backend.create_hardware_metric.assert_has_calls(
            [
                call(
                    experiment,
                    Metric(
                        internal_id=cpu_metric_id,
                        name=u"CPU - usage",
                        description=u"average of all cores",
                        resource_type=MetricResourceType.CPU,
                        unit=u"%",
                        min_value=0.0,
                        max_value=100.0,
                        gauges=[SystemCpuUsageGauge()],
                    ),
                ),
                call(
                    experiment,
                    Metric(
                        internal_id=ram_metric_id,
                        name=u"RAM",
                        description=u"",
                        resource_type=MetricResourceType.RAM,
                        unit=u"GB",
                        min_value=0.0,
                        max_value=memory_amount_gb,
                        gauges=[SystemMemoryUsageGauge()],
                    ),
                ),
            ]
        )

    @unittest.skipIf(IS_MACOS, "MacOS behaves strangely")
    def test_report_and_send_metrics(self):
        # given
        experiment_start = time.time()
        second_after_start = experiment_start + 1.0

        # and
        experiment = MagicMock()

        # and
        metric_service = self.metric_service_factory.create(
            gauge_mode=GaugeMode.SYSTEM,
            experiment=experiment,
            reference_timestamp=experiment_start,
        )
        metrics_container = metric_service.metrics_container

        # when
        metric_service.report_and_send(timestamp=second_after_start)

        # then
        self.backend.send_hardware_metric_reports.assert_called_once_with(
            experiment,
            metrics_container.metrics(),
            [
                MetricReport(
                    metric=metrics_container.cpu_usage_metric,
                    values=[
                        MetricValue(
                            timestamp=second_after_start,
                            running_time=1.0,
                            gauge_name=u"cpu",
                            value=ANY,
                        )
                    ],
                ),
                MetricReport(
                    metric=metrics_container.memory_metric,
                    values=[
                        MetricValue(
                            timestamp=second_after_start,
                            running_time=1.0,
                            gauge_name=u"ram",
                            value=ANY,
                        )
                    ],
                ),
            ],
        )


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.exceptions import NeptuneInternalException


class ContainerTypeFile:
    FILE_NAME = "container_type"

    def __init__(self, dir_path: Path, expected_container_type: ContainerType = None):
        self._file = dir_path / self.FILE_NAME
        self._container_type = self.check_container_type(expected_container_type)

    @property
    def container_type(self) -> ContainerType:
        return self._container_type

    def check_container_type(
        self, expected_container_type: ContainerType
    ) -> ContainerType:
        """Make sure that queue will serve requested `default_container_type`
        or analyze container_type based on information stored on disk."""
        container_type_from_file = None
        if self._file.exists():
            with open(self._file, "r") as f:
                container_type_from_file = ContainerType(f.read())

        if container_type_from_file is None:
            # No information about type stored on disk
            if expected_container_type is None:
                # When working with legacy data
                # container_type couldn't be determined
                # and for backward compatibility
                # it should be Run at default
                return ContainerType.RUN
            return expected_container_type
        else:
            # Information about type is stored on disk
            if (
                expected_container_type is not None
                and container_type_from_file != expected_container_type
            ):
                raise NeptuneInternalException(
                    f"Expected container_type ({expected_container_type.value})"
                    f" doesn't match the one from file ({container_type_from_file.value})"
                )
            return container_type_from_file

    def save(self):
        """Saves information regarding container_type in queue directory"""
        with open(self._file, "w") as f:
            f.write(self._container_type.value)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import random
import concurrent.futures

import pytest
from faker import Faker
import neptune.new as neptune

from e2e_tests.base import BaseE2ETest

fake = Faker()


def store_in_run(run_short_id: str, project: str, destination: str):
    reinitialized_run = neptune.init(run=run_short_id, project=project)
    reinitialized_run[destination] = fake.color()
    reinitialized_run.sync()


class TestMultipleRuns(BaseE2ETest):
    @pytest.mark.parametrize("container", ["run"], indirect=True)
    def test_multiple_runs_single(self, container: neptune.Run, environment):
        # pylint: disable=protected-access,undefined-loop-variable

        number_of_reinitialized = 5
        namespace = fake.unique.word()

        reinitialized_runs = [
            neptune.init(run=container._short_id, project=environment.project)
            for _ in range(number_of_reinitialized)
        ]

        container[f"{namespace}/{fake.unique.word()}"] = fake.color()
        container.sync()

        random.shuffle(reinitialized_runs)
        for run in reinitialized_runs:
            run[f"{namespace}/{fake.unique.word()}"] = fake.color()

        random.shuffle(reinitialized_runs)
        for run in reinitialized_runs:
            run.sync()

        container.sync()

        assert len(container[namespace].fetch()) == number_of_reinitialized + 1

    @pytest.mark.skip(reason="no way of currently testing this")
    @pytest.mark.parametrize("container", ["run"], indirect=True)
    def test_multiple_runs_processes(self, container: neptune.Run, environment):
        # pylint: disable=protected-access

        number_of_reinitialized = 10
        namespace = fake.unique.word()

        container[f"{namespace}/{fake.unique.word()}"] = fake.color()

        with concurrent.futures.ProcessPoolExecutor(max_workers=3) as executor:
            futures = [
                executor.submit(
                    store_in_run,
                    container._short_id,
                    environment.project,
                    f"{namespace}/{fake.unique.word()}",
                )
                for _ in range(number_of_reinitialized)
            ]
            for future in concurrent.futures.as_completed(futures):
                _ = future.result()

        container.sync()

        assert len(container[namespace].fetch()) == number_of_reinitialized + 1

    @pytest.mark.parametrize("container", ["run"], indirect=True)
    def test_multiple_runs_thread(self, container: neptune.Run, environment):
        # pylint: disable=protected-access

        number_of_reinitialized = 10
        namespace = fake.unique.word()

        container[f"{namespace}/{fake.unique.word()}"] = fake.color()

        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            futures = [
                executor.submit(
                    store_in_run,
                    container._short_id,
                    environment.project,
                    f"{namespace}/{fake.unique.word()}",
                )
                for _ in range(number_of_reinitialized)
            ]
            for future in concurrent.futures.as_completed(futures):
                _ = future.result()

        container.sync()

        assert len(container[namespace].fetch()) == number_of_reinitialized + 1



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from datetime import datetime

import mock

from neptune.internal.streams.channel_writer import ChannelWriter


class TestChannelWriter(unittest.TestCase):
    def test_write_data_to_channel_writer(self):
        # given
        experiment = mock.MagicMock()
        experiment.get_system_properties.return_value = {"created": datetime.now()}
        channel_name = "a channel name"
        writer = ChannelWriter(experiment, channel_name)

        # when
        writer.write("some\ndata")

        # then
        # pylint: disable=protected-access
        experiment._channels_values_sender.send.assert_called_once()

    @mock.patch("neptune.internal.streams.channel_writer.datetime")
    def test_write_data_with_low_resolution_datetime_now(self, dt):
        # given
        experiment = mock.MagicMock()
        experiment.get_system_properties.return_value = {
            "created": datetime(2022, 2, 2, 2, 2, 2, 2)
        }
        channel_name = "a channel name"
        writer = ChannelWriter(experiment, channel_name)

        # and
        dt.now.return_value = datetime(2022, 2, 2, 2, 2, 2, 3)

        # when
        writer.write("text1\ntext2\n")

        # then
        # pylint: disable=protected-access
        x_to_text = self._extract_x_to_text_from_calls(
            experiment._channels_values_sender.send.call_args_list
        )
        self.assertEqual(x_to_text, {0.001: "text1", 0.002: "text2"})

    @mock.patch("neptune.internal.streams.channel_writer.datetime")
    def test_write_data_with_high_resolution_datetime_now(self, dt):
        # given
        experiment = mock.MagicMock()
        experiment.get_system_properties.return_value = {
            "created": datetime(2022, 2, 2, 2, 2, 2, 2)
        }
        channel_name = "a channel name"
        writer = ChannelWriter(experiment, channel_name)

        # when
        dt.now.return_value = datetime(2022, 2, 2, 2, 2, 2, 4)
        writer.write("text1\n")
        dt.now.return_value = datetime(2022, 2, 2, 2, 2, 2, 5)
        writer.write("text2\n")

        # then
        # pylint: disable=protected-access
        x_to_text = self._extract_x_to_text_from_calls(
            experiment._channels_values_sender.send.call_args_list
        )
        self.assertEqual(x_to_text, {0.002: "text1", 0.003: "text2"})

    @staticmethod
    def _extract_x_to_text_from_calls(calls):
        channel_values = [kwargs["channel_value"] for (_, kwargs) in calls]
        return dict((v.x, v.y["text_value"]) for v in channel_values)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2021, self.experiment Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remember to set environment values:
* self.experiment_API_TOKEN
* self.experiment_PROJECT
"""
import os
import sys
from datetime import datetime

from neptune import Session, envs
from common_client_code import ClientFeatures


class OldClientNonglobalFeatures(ClientFeatures):
    def __init__(self):
        super().__init__()
        self.session = Session()
        self.project = self.session.get_project(os.getenv(envs.PROJECT_ENV_NAME))
        self.experiment = self.project.create_experiment(
            name="const project name",
            params=self.params,
            tags=["initial tag 1", "initial tag 2"],
        )

    def modify_tags(self):
        self.experiment.append_tags("tag1")
        self.experiment.append_tag(["tag2_to_remove", "tag3"])
        # self.experiment.remove_tag('tag2_to_remove')  # TODO: NPT-9222
        # self.experiment.remove_tag('tag4_remove_non_existing')  # TODO: NPT-9222

        exp = self.experiment
        assert set(exp.get_tags()) == {
            "initial tag 1",
            "initial tag 2",
            "tag1",
            "tag2_to_remove",
            "tag3",
        }

    def modify_properties(self):
        self.experiment.set_property("prop", "some text")
        self.experiment.set_property("prop_number", 42)
        self.experiment.set_property("nested/prop", 42)
        self.experiment.set_property("prop_to_del", 42)
        self.experiment.set_property("prop_list", [1, 2, 3])
        with open(self.text_file_path, mode="r") as f:
            self.experiment.set_property("prop_IO", f)
        self.experiment.set_property("prop_datetime", datetime.now())
        self.experiment.remove_property("prop_to_del")

        properties = self.experiment.get_properties()
        assert properties["prop"] == "some text"
        assert properties["prop_number"] == "42"
        assert properties["nested/prop"] == "42"
        assert "prop_to_del" not in properties
        assert (
            properties["prop_IO"]
            == "<_io.TextIOWrapper name='alpha_integration_dev/data/text.txt'"
            " mode='r' encoding='UTF-8'>"
        )
        print(f"Properties: {properties}")

    def log_std(self):
        print("stdout text1")
        print("stdout text2")
        print("stderr text1", file=sys.stderr)
        print("stderr text2", file=sys.stderr)

    def log_series(self):
        # floats
        self.experiment.log_metric("m1", 1)
        self.experiment.log_metric("m1", 2)
        self.experiment.log_metric("m1", 3)
        self.experiment.log_metric("m1", 2)
        self.experiment.log_metric("nested/m1", 1)

        # texts
        self.experiment.log_text("m2", "a")
        self.experiment.log_text("m2", "b")
        self.experiment.log_text("m2", "c")

        # images
        # `image_name` and `description` will be lost
        self.experiment.log_image(
            "g_img", self.img_path, image_name="name", description="desc"
        )
        self.experiment.log_image("g_img", self.img_path)

        # see what we've logged
        logs = self.experiment.get_logs()
        print(f"Logs: {logs}")

    def handle_files_and_images(self):
        # image
        # `image_name` and `description` will be lost (`send_image` the same as `log_image`)
        self.experiment.send_image(
            "image", self.img_path, name="name", description="desc"
        )

        # artifact
        # (`log_artifact` the same as `log_artifact`)
        self.experiment.send_artifact(self.text_file_path)
        self.experiment.log_artifact(
            self.text_file_path, destination="dir/text file artifact"
        )
        with open(self.text_file_path, mode="r") as f:
            self.experiment.send_artifact(f, destination="file stream.txt")
        self.experiment.log_artifact(self.img_path, destination="dir to delete/art1")
        self.experiment.log_artifact(self.img_path, destination="dir to delete/art2")
        # self.experiment.delete_artifacts('dir to delete')  # doesn't work for alpha NPT-9250
        self.experiment.delete_artifacts("dir to delete/art1")

    def finalize(self):
        pass


if __name__ == "__main__":
    OldClientNonglobalFeatures().run()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from io import BytesIO
from mock import patch

from neptune.internal.storage.datastream import FileChunkStream, FileChunk
from neptune.internal.storage.storage_utils import (
    UploadEntry,
    AttributeUploadConfiguration,
)


class TestFileChunkStream(unittest.TestCase):
    @patch("os.path.exists", new=lambda _: True)
    @patch("stat.S_ISDIR", new=lambda _: False)
    @patch("os.lstat")
    def test_permissions_to_unix_string_for_file(self, lstat):
        # given
        lstat.return_value.st_mode = 0o731

        # when
        permissions_string = UploadEntry.permissions_to_unix_string("/some/path")

        # then
        self.assertEqual("-rwx-wx--x", permissions_string)

    @patch("os.path.exists", new=lambda _: True)
    @patch("stat.S_ISDIR", new=lambda _: True)
    @patch("os.lstat")
    def test_permissions_to_unix_string_for_directory(self, lstat):
        # given
        lstat.return_value.st_mode = 0o642

        # when
        permissions_string = UploadEntry.permissions_to_unix_string("/some/path")

        # then
        self.assertEqual("drw-r---w-", permissions_string)

    @patch("os.path.exists", new=lambda _: False)
    def test_permissions_to_unix_string_for_nonexistent_file(self):
        # when
        permissions_string = UploadEntry.permissions_to_unix_string("/some/path")

        # then
        self.assertEqual("-" * 10, permissions_string)

    def test_generate_chunks_from_stream(self):
        # given
        text = u"ABCDEFGHIJKLMNOPRSTUWXYZ"

        # when
        stream = FileChunkStream(
            UploadEntry(BytesIO(bytes(text, "utf-8")), "some/path"),
            AttributeUploadConfiguration(10),
        )
        chunks = list()
        for chunk in stream.generate():
            chunks.append(chunk)

        # then
        self.assertEqual(stream.length, 24)
        self.assertEqual(
            chunks,
            [
                FileChunk(b"ABCDEFGHIJ", 0, 10),
                FileChunk(b"KLMNOPRSTU", 10, 20),
                FileChunk(b"WXYZ", 20, 24),
            ],
        )


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import unittest
from io import StringIO
from contextlib import redirect_stdout
from unittest.mock import MagicMock

from neptune.new.internal.streams.std_stream_capture_logger import StdoutCaptureLogger


class TestStdStreamCaptureLogger(unittest.TestCase):
    def test_catches_stdout(self):
        stdout = StringIO()
        with redirect_stdout(stdout):
            mock_run = MagicMock()
            attr_name = "sys/stdout"
            logger = StdoutCaptureLogger(mock_run, attr_name)
            stdout_fp = sys.stdout
            print("testing", file=stdout_fp)
            logger.close()

            self.assertListEqual(
                mock_run[attr_name].log.call_args_list,
                [
                    (("testing",), {}),
                    (("\n",), {}),
                ],
            )
        stdout.seek(0)
        self.assertEqual(stdout.read(), "testing\n")

    def test_does_not_report_if_used_after_stop(self):
        stdout = StringIO()
        with redirect_stdout(stdout):
            mock_run = MagicMock()
            attr_name = "sys/stdout"
            logger = StdoutCaptureLogger(mock_run, attr_name)
            stdout_fp = sys.stdout
            logger.close()

            print("testing", file=stdout_fp)
            mock_run[attr_name].log.assert_not_called()
        stdout.seek(0)
        self.assertEqual(stdout.read(), "testing\n")



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
import logging
import os
import socket
import sys
import time
from functools import lru_cache, wraps
from typing import (
    Optional,
    Dict,
    TYPE_CHECKING,
    Mapping,
    Text,
    Any,
    List,
    Iterable,
)
from urllib.parse import urlparse, urljoin


import click
import requests
import urllib3
from bravado.requests_client import RequestsResponseAdapter
from urllib3.exceptions import NewConnectionError
from bravado.client import SwaggerClient
from bravado.exception import (
    BravadoConnectionError,
    BravadoTimeoutError,
    HTTPForbidden,
    HTTPServerError,
    HTTPUnauthorized,
    HTTPServiceUnavailable,
    HTTPRequestTimeout,
    HTTPGatewayTimeout,
    HTTPBadGateway,
    HTTPClientError,
    HTTPTooManyRequests,
    HTTPError,
)
from bravado.http_client import HttpClient
from bravado_core.formatter import SwaggerFormat
from packaging.version import Version
from requests import Session, Response

from neptune.new.envs import (
    NEPTUNE_RETRIES_TIMEOUT_ENV,
    NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE,
)
from neptune.new.exceptions import (
    SSLError,
    NeptuneConnectionLostException,
    Unauthorized,
    Forbidden,
    CannotResolveHostname,
    UnsupportedClientVersion,
    ClientHttpError,
    NeptuneFeatureNotAvailableException,
    MetadataInconsistency,
    NeptuneInvalidApiTokenException,
)
from neptune.new.internal.backends.api_model import ClientConfig
from neptune.new.internal.operation import Operation, CopyAttribute
from neptune.new.internal.utils import replace_patch_version

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from neptune.new.internal.backends.neptune_backend import NeptuneBackend
    from neptune.new.internal.backends.hosted_neptune_backend import (
        HostedNeptuneBackend,
    )

MAX_RETRY_TIME = 30
retries_timeout = int(os.getenv(NEPTUNE_RETRIES_TIMEOUT_ENV, "60"))


def with_api_exceptions_handler(func):
    def wrapper(*args, **kwargs):
        last_exception = None
        start_time = time.monotonic()
        for retry in itertools.count(0):
            if time.monotonic() - start_time > retries_timeout:
                break

            try:
                return func(*args, **kwargs)
            except requests.exceptions.InvalidHeader as e:
                if "X-Neptune-Api-Token" in e.args[0]:
                    raise NeptuneInvalidApiTokenException()
                raise
            except requests.exceptions.SSLError as e:
                raise SSLError() from e
            except (
                BravadoConnectionError,
                BravadoTimeoutError,
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                HTTPRequestTimeout,
                HTTPServiceUnavailable,
                HTTPGatewayTimeout,
                HTTPBadGateway,
                HTTPTooManyRequests,
                HTTPServerError,
                NewConnectionError,
            ) as e:
                time.sleep(min(2 ** min(10, retry), MAX_RETRY_TIME))
                last_exception = e
                continue
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except HTTPClientError as e:
                raise ClientHttpError(e.status_code, e.response.text) from e
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code in (
                    HTTPRequestTimeout.status_code,
                    HTTPBadGateway.status_code,
                    HTTPServiceUnavailable.status_code,
                    HTTPGatewayTimeout.status_code,
                    HTTPTooManyRequests.status_code,
                    HTTPServerError.status_code,
                ):
                    time.sleep(min(2 ** min(10, retry), MAX_RETRY_TIME))
                    last_exception = e
                    continue
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                elif 400 <= status_code < 500:
                    raise ClientHttpError(status_code, e.response.text) from e
                else:
                    raise
        raise NeptuneConnectionLostException(last_exception) from last_exception

    return wrapper


@lru_cache(maxsize=None, typed=True)
def verify_host_resolution(url: str) -> None:
    host = urlparse(url).netloc.split(":")[0]
    try:
        socket.gethostbyname(host)
    except socket.gaierror:
        raise CannotResolveHostname(host)


uuid_format = SwaggerFormat(
    format="uuid",
    to_python=lambda x: x,
    to_wire=lambda x: x,
    validate=lambda x: None,
    description="",
)


@with_api_exceptions_handler
def create_swagger_client(url: str, http_client: HttpClient) -> SwaggerClient:
    return SwaggerClient.from_url(
        url,
        config=dict(
            validate_swagger_spec=False,
            validate_requests=False,
            validate_responses=False,
            formats=[uuid_format],
        ),
        http_client=http_client,
    )


def verify_client_version(client_config: ClientConfig, version: Version):
    version_with_patch_0 = Version(replace_patch_version(str(version)))
    if (
        client_config.version_info.min_compatible
        and client_config.version_info.min_compatible > version
    ):
        raise UnsupportedClientVersion(
            version, min_version=client_config.version_info.min_compatible
        )
    if (
        client_config.version_info.max_compatible
        and client_config.version_info.max_compatible < version_with_patch_0
    ):
        raise UnsupportedClientVersion(
            version, max_version=client_config.version_info.max_compatible
        )
    if (
        client_config.version_info.min_recommended
        and client_config.version_info.min_recommended > version
    ):
        click.echo(
            "WARNING: We recommend an upgrade to a new version of neptune-client - {} (installed - {}).".format(
                client_config.version_info.min_recommended, version
            ),
            sys.stderr,
        )


def update_session_proxies(session: Session, proxies: Optional[Dict[str, str]]):
    if proxies:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError("Wrong proxies format: {}".format(proxies))


def build_operation_url(base_api: str, operation_url: str) -> str:
    if "://" not in base_api:
        base_api = f"https://{base_api}"

    return urljoin(base=base_api, url=operation_url)


# TODO print in color once colored exceptions are added
def handle_server_raw_response_messages(response: Response):
    try:
        info = response.headers.get("X-Server-Info")
        if info:
            click.echo(info)
        warning = response.headers.get("X-Server-Warning")
        if warning:
            click.echo(warning)
        error = response.headers.get("X-Server-Error")
        if error:
            click.echo(message=error, err=True)
        return response
    except Exception:
        # any issues with printing server messages should not cause code to fail
        return response


# TODO print in color once colored exceptions are added
class NeptuneResponseAdapter(RequestsResponseAdapter):
    @property
    def raw_bytes(self) -> bytes:
        self._handle_response()
        return super().raw_bytes

    @property
    def text(self) -> Text:
        self._handle_response()
        return super().text

    def json(self, **kwargs) -> Mapping[Text, Any]:
        self._handle_response()
        return super().json(**kwargs)

    def _handle_response(self):
        try:
            info = self._delegate.headers.get("X-Server-Info")
            if info:
                click.echo(info)
            warning = self._delegate.headers.get("X-Server-Warning")
            if warning:
                click.echo(warning)
            error = self._delegate.headers.get("X-Server-Error")
            if error:
                click.echo(message=error, err=True)
        except Exception:
            # any issues with printing server messages should not cause code to fail
            pass


class MissingApiClient(SwaggerClient):
    """catch-all class to gracefully handle calls to unavailable API"""

    def __init__(self, feature_name: str):  # pylint: disable=super-init-not-called
        self.feature_name = feature_name

    def __getattr__(self, item):
        raise NeptuneFeatureNotAvailableException(missing_feature=self.feature_name)


# https://stackoverflow.com/a/44776960
def cache(func):
    """
    Transform mutable dictionary into immutable before call to lru_cache
    """

    class HDict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

    func = lru_cache(maxsize=None, typed=True)(func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        args = tuple([HDict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: HDict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)

    wrapper.cache_clear = func.cache_clear
    return wrapper


def ssl_verify():
    if os.getenv(NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE):
        urllib3.disable_warnings()
        return False

    return True


def parse_validation_errors(error: HTTPError) -> Dict[str, str]:
    return {
        f"{error_description.get('errorCode').get('name')}": error_description.get(
            "context", ""
        )
        for validation_error in error.swagger_result.validationErrors
        for error_description in validation_error.get("errors")
    }


class ExecuteOperationsBatchingManager:
    def __init__(self, backend: "NeptuneBackend"):
        self._backend = backend

    def get_batch(
        self, ops: Iterable[Operation], errors: List[MetadataInconsistency]
    ) -> List[Operation]:
        batch = []
        for op in ops:
            if isinstance(op, CopyAttribute):
                if not batch:
                    try:
                        # CopyAttribute can be at the start of a batch
                        batch.append(op.resolve(self._backend))
                    except MetadataInconsistency as e:
                        errors.append(e)
                else:
                    # cannot have CopyAttribute after any other op in a batch
                    break
            else:
                batch.append(op)

        return batch
